/**
 * This source code is a component of Charlie Feng's source library,
 * All copyright reverved.
 * @author Charlie Feng (Fengertao@hotmail.com)
 */
package charlie.feng.xml.XML2Table;


import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import charlie.feng.sql.datasource.JDBCODBCConnectionFactory;



/**
 * Some Concept Define by me
 * Simple Column: like below
 * 			<Type>LEADER_BARBARIAN</Type>
 * In DB the column "Type" will have value "LEADER_BARBARIAN"
 * 
 * 
 * Composit Column: like blow
 * 			<Traits>
				<Trait>
					<TraitType>TRAIT_PHILOSOPHICAL</TraitType>
					<bTrait>1</bTrait>
				</Trait>
				<Trait>
					<TraitType>TRAIT_AGGRESSIVE</TraitType>
					<bTrait>1</bTrait>
				</Trait>
			</Traits>
 
 * 
 * In DB there will have 2 "Trait" column, first column have value "TRAIT_PHILOSOPHICAL:1"
 * second column have value "TRAIT_AGGRESSIVE:1"
 * 
 * @author Charlie Feng
 */
public abstract class PlainXML2Table {

	//blow are hard coded constant
	static int COLUMN_LENGTH = 50;				//hard code column length to 50
	static int COMPOSIT_COLUMN_LENGTH = 100;	//acturely useless because I hard to all to 50
	static int NULL_COLUMN_LENGTH = 5;			//acturely useless because I hard to all to 50
	static String COMPOSIT_VALUE = "**";		//if I don't want display some composit column, display as "**" 
	static String EMPTY_VALUE = "";				//value for EMPTY column
	static String DILIMITER = ":";				//dilimiter for composit column.

	//below are config info, define in child class.	
	String xmlPath;
	String dbUrl;
	String id;
	String pwd;
	String recordTag;						//the Eye-catcher of record start.
	String tableName;						//the table name of excel (i.e. the sheet name)
	Map compositTagMap = new HashMap();		//Key:tags contains composit tag which need display.
											//Value: maximam number. 
	Set escapeTagSet = new HashSet();		//tags needn't display
	
	int colSn = 0;
	Map columnMap = new HashMap();
	String insertHeaderSql = "";
	Connection con;
	
	public void run() throws Exception{
		try{
			Class.forName("sun.jdbc.odbc.JdbcOdbcDriver");
			con = DriverManager.getConnection(dbUrl, id, pwd);
			con.setAutoCommit(true);
			
			DocumentBuilderFactory dbf =  DocumentBuilderFactory.newInstance();
			DocumentBuilder builder = dbf.newDocumentBuilder();
			dbf.setValidating(true);
			Document document = builder.parse(xmlPath);
			Element element = document.getDocumentElement();
			buildColumnMap(element);
			createTable();
			insertRecords(element);
			System.out.println("Ended.");

		} catch (Exception ex) {
			ex.printStackTrace();
		} finally {
			if (con != null) {
				con.close();
			}
		}
		
	}
	
	public void buildColumnMap(Node node) {
		if (node == null) {
			return;
		}
		
		if (node.getNodeType() != node.ELEMENT_NODE) {
			return;
		}
		 
		if (!(node.getNodeName().equals(recordTag))) {
			NodeList children = node.getChildNodes();
			if (null == children) {
				return;
			}
			for ( int i = 0; i< children.getLength(); i++) {
				buildColumnMap(children.item(i));
			}
			return;
		} else {
			renewColumnMap(node);
		}
	}
	
	public void renewColumnMap(Node node) {
		NodeList children = node.getChildNodes();
		if (null == children) {
			return;
		}
columnConfigLoop:
		for ( int i = 0; i< children.getLength(); i++) {
			Node columnNode = children.item(i);
			
			if (columnNode.getNodeType() == Node.ELEMENT_NODE) {
				String columnDesc = columnNode.getNodeName();
				int columnLength = 0;
				if (escapeTagSet.contains(columnDesc)) {
					continue columnConfigLoop;
				}
				if (compositTagMap.get(columnDesc) != null) {
					configCompositColumn(columnNode);
					continue columnConfigLoop;
				}
				
				NodeList columnChildren = columnNode.getChildNodes();
				if (columnChildren.getLength() == 0) {
					columnLength = 0;
				} else if (columnChildren.getLength() == 1){
					String value = columnChildren.item(0).getNodeValue();
					if (value == null) {
						columnLength = NULL_COLUMN_LENGTH;
					} else {
						columnLength = value.length();
					}
				} else {
					columnLength = COMPOSIT_COLUMN_LENGTH;
				}
				Column column = (Column)columnMap.get(columnDesc); 
				if (column == null) {
					column = new Column(geneColSN(1), columnDesc, Math.max(columnDesc.length(), columnLength));
				} else {
					column.setLength(Math.max(column.getLength(), columnLength));
					
				}
				columnMap.put(columnDesc,column);
			}
		}
		return;
	}
	
	
	public void configCompositColumn(Node node) {
		String columnsDesc = node.getNodeName();
		if (compositTagMap.get(columnsDesc) != null) {
			int columnCount = ((Integer)compositTagMap.get(columnsDesc)).intValue();
			Column column = (Column)columnMap.get(columnsDesc); 
			if (column == null) {
				column = new Column(geneColSN(columnCount), columnsDesc, Math.max(columnsDesc.length()+2, COLUMN_LENGTH), columnCount);
			} else {
				column.setLength(Math.max(column.getLength(), COLUMN_LENGTH));
			}
			columnMap.put(columnsDesc,column);
		}
	}
	
	public void insertCompositColumn(Node node, PreparedStatement ps) throws SQLException{
		String columnsDesc = node.getNodeName();
		if (compositTagMap.get(columnsDesc) != null) {
			Column column = (Column)columnMap.get(columnsDesc);
			int maxCount = column.getCount();
			int leadSn = column.getSn();
			NodeList nodelist = node.getChildNodes();
			int innerSn = 0;
			for (int i = 0; i < nodelist.getLength(); i++) {
				Node singleColumn = nodelist.item(i);
				String singleColumnValue = EMPTY_VALUE;
				if (singleColumn.getNodeType() == Node.ELEMENT_NODE) {
					NodeList columnPartNodeList = singleColumn.getChildNodes();
					for (int j = 0; j < columnPartNodeList.getLength(); j++) {
						Node columnPartNode = columnPartNodeList.item(j);
						if (columnPartNode.getNodeType() == Node.ELEMENT_NODE) {
							NodeList columnPartTextNodeList = columnPartNode.getChildNodes();
							for (int k = 0; k < columnPartTextNodeList.getLength(); k++){
								Node columnPartTextNode = columnPartTextNodeList.item(k);
								if (columnPartTextNode.getNodeType() == Node.TEXT_NODE) {
									singleColumnValue += columnPartTextNode.getNodeValue() + DILIMITER;
								}
							}
						}
					}
					if (singleColumnValue.length() > 0) {
						singleColumnValue = singleColumnValue.substring(0, singleColumnValue.length() - 1);
					}
					ps.setString(leadSn + innerSn, singleColumnValue);
					innerSn ++;
				}
			}
			
			for (int j = innerSn; j < maxCount; j++) {
				ps.setString(leadSn + j, EMPTY_VALUE);
			}
			 
		}
	}
	
	
	public void createTable() throws SQLException{
		String createTableSql = "create table " + tableName + "( ";
		for (int i = 1; i <= colSn; i++) {
			Integer columnSn = new Integer(i);
//			String columnDesc = (String)columnSnMap.get(columnSn);
//			Column column = (Column)columnMap.get(columnDesc);
			createTableSql += "COL" + i;
			createTableSql += " varchar(";
//			createTableSql += column.getLength();
			createTableSql += COLUMN_LENGTH;
			createTableSql += ")";
			if (i < colSn) {
				createTableSql += ",";
			}
		}
		createTableSql += ")";
		System.out.println("createTableSql:" + createTableSql);
		Statement stat = con.createStatement();
		stat.executeUpdate(createTableSql);
		stat.close();
		
		String insertHeaderSql1 = "insert into " + tableName + "( ";
		String insertHeaderSql2 = ") values (";
		String insertHeaderSql3 = ")";
		
		for (int i = 1; i<=colSn; i++) {
			insertHeaderSql1 += "COL" + i;
			insertHeaderSql2 += "?";
			
			if (i < colSn) {
				insertHeaderSql1 += ",";
				insertHeaderSql2 += ",";
			}
		}
		insertHeaderSql = insertHeaderSql1 + insertHeaderSql2 + insertHeaderSql3;
		System.out.println("insertHeaderSql:" + insertHeaderSql);
		
		PreparedStatement ps = con.prepareStatement(insertHeaderSql) ;
		for (Iterator iter = columnMap.entrySet().iterator(); iter.hasNext(); ) {
			Map.Entry entry = (Map.Entry)iter.next();
			String columnDesc = (String)entry.getKey();
			Column column = (Column)entry.getValue();
			int leadSn = column.getSn();
			int count = column.getCount();
			String desc = column.getDesc();
			for (int i = 0; i < count; i++) {
				ps.setString(leadSn + i, desc);
							
			}
			
		}
		ps.executeUpdate();
		ps.close();
	}
	
	public void insertRecords(Node node) throws SQLException{
		if (node == null) {
			return;
		}
		
		if (node.getNodeType() != node.ELEMENT_NODE) {
			return;
		}
		
		if (!(node.getNodeName().equals(recordTag))) {
			NodeList children = node.getChildNodes();
			if (null == children) {
				return;
			}
			for ( int i = 0; i< children.getLength(); i++) {
				insertRecords(children.item(i));
			}
			return;
		} else {
			insertRecord(node);
		}
	}
	
	public void insertRecord(Node node) throws SQLException{
		PreparedStatement ps = con.prepareStatement(insertHeaderSql) ;
		NodeList children = node.getChildNodes();
		if (null == children) {
			return;
		}
columnInsertLoop:
			for ( int i = 0; i< children.getLength(); i++) {
			Node columnNode = children.item(i);
			if (columnNode.getNodeType() == Node.ELEMENT_NODE) {
				String columnDesc = columnNode.getNodeName();
				String columnValue = EMPTY_VALUE;
				if (escapeTagSet.contains(columnDesc)) {
					continue columnInsertLoop;
				}
				if (compositTagMap.get(columnDesc) != null) {
					insertCompositColumn(columnNode, ps);
					continue columnInsertLoop;
				}
				
				NodeList columnChildren = columnNode.getChildNodes();
				if (columnChildren.getLength() == 0) {
					columnValue = EMPTY_VALUE;
				} else if (columnChildren.getLength() == 1){
					columnValue = columnChildren.item(0).getNodeValue();
				} else {
					columnValue = COMPOSIT_VALUE;
				}
				int columnSn = ((Column)columnMap.get(columnDesc)).getSn();
				ps.setString(columnSn, columnValue);
			}
			
		}
		ps.executeUpdate();
		ps.close();
		return;
	}

	public int geneColSN( int snUsed) {
		int firstSn = colSn + 1;
		colSn += snUsed;
		return firstSn;
	}
}
